from __future__ import absolute_import
from vtkmodules import vtkFiltersAMR
from vtkmodules import vtkFiltersExtraction
from vtkmodules import vtkFiltersGeneral
from vtkmodules import vtkFiltersParallelStatistics
from vtkmodules import vtkIOEnSight
from vtkmodules import vtkIOImage
from vtkmodules import vtkIOImport
from vtkmodules import vtkIOParallel
from vtkmodules import vtkIOParallelExodus
from vtkmodules import vtkIOParallelXML
from vtkmodules import vtkImagingFourier
from vtkmodules import vtkImagingSources
from vtkmodules import vtkInteractionWidgets
from . import vtkPVVTKExtensionsCore
from . import vtkPVVTKExtensionsRendering
from .vtkPVVTKExtensionsDefaultPython import *



import paraview

try:
    from vtkmodules.vtkCommonComputationalGeometry import *
except ImportError:
    paraview.print_error("Error: Could not import vtkCommonComputationalGeometry")
from vtkmodules.vtkCommonCore import *
from vtkmodules.vtkCommonDataModel import *
from vtkmodules.vtkCommonExecutionModel import *
try:
    from vtkmodules.vtkCommonMath import *
except ImportError:
    paraview.print_error("Error: Could not import vtkCommonMath")
try:
    from vtkmodules.vtkCommonMisc import *
except ImportError:
    paraview.print_error("Error: Could not import vtkCommonMisc")
try:
    from vtkmodules.vtkCommonSystem import *
except ImportError:
    paraview.print_error("Error: Could not import vtkCommonSystem")
try:
    from vtkmodules.vtkCommonTransforms import *
except ImportError:
    paraview.print_error("Error: Could not import vtkCommonTransforms")
from vtkmodules.vtkFiltersProgrammable import *
from vtkmodules.vtkParallelCore import *
try:
    from vtkmodules.vtkRenderingCore import vtkCamera
except ImportError:
    paraview.print_error("Error: Could not import vtkRenderingCore")
try:
    from vtkmodules.vtkFiltersCore import *
except ImportError:
    paraview.print_error("Error: Could not import vtkFiltersCore")



r"""
    This module is a ParaViewWeb server application used only for automated testing
"""

import os

# import paraview modules.
from paraview.web import pv_wslink
from paraview.web import protocols as pv_protocols
from wslink import server

try:
    import argparse
except ImportError:
    # since  Python 2.6 and earlier don't have argparse, we simply provide
    # the source for the same as _argparse and we use it instead.
    from vtk.util import _argparse as argparse


# =============================================================================
# Create custom Pipeline Manager class to handle clients requests
# =============================================================================

class _TestServer(pv_wslink.PVServerProtocol):

    dataDir = os.getcwd()
    canFile = None
    authKey = "wslink-secret"
    dsHost = None
    dsPort = 11111
    rsHost = None
    rsPort = 11111
    rcPort = -1
    fileToLoad = None
    groupRegex = "[0-9]+\\.[0-9]+\\.|[0-9]+\\."
    excludeRegex = "^\\.|~$|^\\$"
    plugins = None
    filterFile = None
    colorPalette = None
    proxies = None
    allReaders = True
    saveDataDir = os.getcwd()
    viewportScale=1.0
    viewportMaxWidth=2560
    viewportMaxHeight=1440

    def initialize(self):
        # Bring used components
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebStartupRemoteConnection(_TestServer.dsHost, _TestServer.dsPort, _TestServer.rsHost, _TestServer.rsPort, _TestServer.rcPort))
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebStartupPluginLoader(_TestServer.plugins))
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebFileListing(_TestServer.dataDir, "Home", _TestServer.excludeRegex, _TestServer.groupRegex))
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebProxyManager(allowedProxiesFile=_TestServer.proxies, baseDir=_TestServer.dataDir, fileToLoad=_TestServer.fileToLoad, allowUnconfiguredReaders=_TestServer.allReaders))
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebColorManager(pathToColorMaps=_TestServer.colorPalette))
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebMouseHandler())
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebViewPort()) # _TestServer.viewportScale, _TestServer.viewportMaxWidth, _TestServer.viewportMaxHeight
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebViewPortImageDelivery())
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebViewPortGeometryDelivery())
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebTimeHandler())
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebSelectionHandler())
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebWidgetManager())
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebKeyValuePairStore())
        self.registerVtkWebProtocol(pv_protocols.ParaViewWebSaveData(baseSavePath=_TestServer.saveDataDir))

        # Update authentication key to use
        self.updateSecret(_TestServer.authKey)

# =============================================================================
# Main: Parse args and start server
# =============================================================================

if __name__ == "__main__":
    # Create argument parser
    parser = argparse.ArgumentParser(description="ParaView Web Test Server")

    # Add arguments
    server.add_arguments(parser)
    args = parser.parse_args()

    # Start server
    server.start_webserver(options=args, protocol=_TestServer)



from __future__ import absolute_import
from . import vtkCommonCore
from . import vtkCommonDataModel
from . import vtkCommonExecutionModel
from . import vtkCommonMisc
from .vtkFiltersCorePython import *



r"""wslink is a module that extends any
wslink related classes for the purposes of vtkWeb.

"""

from __future__ import absolute_import, division, print_function

# import inspect, types, string, random, logging, six, json, re, base64
import json, base64, time

from twisted.python         import log
from twisted.internet       import reactor

from autobahn.twisted.websocket import WebSocketServerProtocol

from wslink import websocket
from wslink import register as exportRpc

from vtk.web import protocols
from vtk.vtkWebCore import vtkWebApplication

# =============================================================================
application = None

# =============================================================================
#
# Base class for vtkWeb ServerProtocol
#
# =============================================================================

class ServerProtocol(websocket.ServerProtocol):
    """
    Defines the core server protocol for vtkWeb. Adds support to
    marshall/unmarshall RPC callbacks that involve ServerManager proxies as
    arguments or return values.

    Applications typically don't use this class directly, but instead
    sub-class it and call self.registerVtkWebProtocol() with useful vtkWebProtocols.
    """

    def __init__(self):
        log.msg('Creating SP')
        self.setSharedObject("app", self.initApplication())
        websocket.ServerProtocol.__init__(self)

    def initApplication(self):
        """
        Let subclass optionally initialize a custom application in lieu
        of the default vtkWebApplication.
        """
        global application
        if not application:
            application = vtkWebApplication()
        return application

    def setApplication(self, application):
        self.setSharedObject("app", application)

    def getApplication(self):
        return self.getSharedObject("app")

    def registerVtkWebProtocol(self, protocol):
        self.registerLinkProtocol(protocol)

    def getVtkWebProtocols(self):
        return self.getLinkProtocols()



r"""Utility module for easy manipultions of lookup tables.
This module is intended for use with by simple.py.


DEPRECATED: will be removed in future releases of ParaView.
"""
#==============================================================================
#
#  Program:   ParaView
#  Module:    lookuptable.py
#
#  Copyright (c) Kitware, Inc.
#  All rights reserved.
#  See Copyright.txt or http://www.paraview.org/HTML/Copyright.html for details.
#
#     This software is distributed WITHOUT ANY WARRANTY; without even
#     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#     PURPOSE.  See the above copyright notice for more information.
#
#==============================================================================
from __future__ import absolute_import
import os
from math import sqrt

from . import servermanager
from ._colorMaps import getColorMaps

# -----------------------------------------------------------------------------

class _vtkPVLUTData:
  """
  Internal container for ParaView lookup table data.
  Don't use this directly. Use vtkPVLUTReader.
  """
  def __init__(self):
    self.Name=""
    self.Space=""
    self.Values=[]
    self.Coords=[]

  def SetName(self,aName):
    self.Name=aName

  def GetName(self):
    return self.Name

  def SetColorSpace(self,aSpace):
    self.Space=aSpace

  def GetColorSpace(self):
    return self.Space

  def SetRGBValues(self,aValues):
    self.Values=aValues

  def GetRGBValues(self):
    return self.Values

  def SetMapCoordinates(self,aCoords):
    self.Coords=aCoords
    # normalize the coordinates
    # in preparation to map onto
    # an arbitrary scalar range
    nCoords = len(self.Coords)
    minCoord = float(min(self.Coords))
    maxCoord = float(max(self.Coords))
    deltaCoord = maxCoord - minCoord
    if (minCoord>=maxCoord):
      print ('ERROR: in coordinate values')
      return
    i=0
    while i<nCoords:
      self.Coords[i] -= minCoord
      self.Coords[i] /= deltaCoord
      i+=1
    return

  def GetMapCoordinates(self):
    return self.Coords;

  def PrintSelf(self):
    print (self.Name)
    print (self.Space)
    print (self.Values)
    print (self.Coords)


# -----------------------------------------------------------------------------

class vtkPVLUTReader:
  """
  Reader and container for ParaView's XML based lookup tables.
  Once lookup tables are loaded you access them by name. When
  accessing you must provide the array instance, which you may
  get from a pvpython 'Source' type object.

  This reader makes use of ParaView's XML LUT file format with
  one exception - the XML document must be root'ed by an element
  named "ColorMaps". Within the "ColorMaps" element an arbitrary
  number of ParaView's "ColorMap" elements define LUT entries.

  Usage:

  ::

     # at the top of your script
     # create the reader and load LUT's
     lr = lookuptable.vtkPVLUTReader()
     lr.Read('/path/to/luts.xml')
     lr.Print()

     # after you have a pvpython source object, get
     # one of it's arrays.
     srcObj = GetActiveSource()
     array = srcObj.PointData.GetArray('arrayName')

     # create a LUT for the array.
     lut = lr.GetLUT(array,'lutName')

     # set the active array and assign the LUT
     srcObjRep = Show(srcObj)
     srcObjRep.ColorArrayName = 'arrayName'
     srcObjRep.LookupTable = lut

     # finally render to display the result
     Render()

  File Format:

  ::

      <ColorMaps>
          ...
        <ColorMap name="LUTName" space="Lab,RGB,HSV" indexedLookup="true,false">
          <Point x="val" o="val" r="val" g="val" b="val"/>
            ...
          <Point x="val" o="val" r="val" g="val" b="val"/>
          <NaN r="val" g="val" b="val"/>
        </ColorMap>
          ...
        <ColorMap>
          ...
        </ColorMap>
          ...
      </ColorMaps>
  """

  def __init__(self,ns=None):
    self.LUTS={}
    self.DefaultLUT=None
    self.Globals=ns

    defaultColorMaps = getColorMaps()
    if defaultColorMaps:
      self._Read(defaultColorMaps)
    else:
      print('WARNING: default LUTs not found.')
    return

  def Clear(self):
    """
    Clear internal data structures.
    """
    self.LUTS={}
    self.DefaultLUT=None
    return

  def Read(self, aFileName):
    """
    Read in the LUT's defined in the named file. Each
    call to read extends the internal list of LUTs.
    """
    parser=servermanager.vtkPVXMLParser()
    parser.SetFileName(aFileName)
    if (not parser.Parse()):
      print ('ERROR: parsing lut file %s'%(aFileName))
      return
    root=parser.GetRootElement()
    if root.GetName()!='ColorMaps':
      print ('ERROR: parsing LUT file %s'%(aFileName))
      print ('ERROR: root element must be <ColorMaps>')
      return
    return self._Read(root)

  def _Read(self, root):
    nElems=root.GetNumberOfNestedElements()
    i=0
    nFound=0
    while (i<nElems):
      cmapElem=root.GetNestedElement(i)
      if (cmapElem.GetName()=='ColorMap'):
        nFound+=1

        lut=_vtkPVLUTData()
        lut.SetName(cmapElem.GetAttribute('name'))
        lut.SetColorSpace(cmapElem.GetAttribute('space'))

        coords=[]
        values=[]
        nRGB=cmapElem.GetNumberOfNestedElements()
        j=0
        while (j<nRGB):
          rgbElem=cmapElem.GetNestedElement(j)
          if (rgbElem.GetName()=='Point'):
            coord=float(rgbElem.GetAttribute('x'))
            coords.append(coord)
            val=[float(rgbElem.GetAttribute('r')),
              float(rgbElem.GetAttribute('g')),
              float(rgbElem.GetAttribute('b'))]
            values.append(val)
          j=j+1

        lut.SetMapCoordinates(coords)
        lut.SetRGBValues(values)
        #lut.PrintSelf()

        self.LUTS[lut.GetName()]=lut
      i=i+1

    if nFound==0:
      print ('ERROR: No ColorMaps were found in %s'%(aFileName))
    else:
      if self.DefaultLUT is None:
        names=list(self.LUTS)
        if len(names)>0:
          self.DefaultLUT=names[0]

    return nFound

  def GetLUT(self,aArray,aLutName,aRangeOveride=[]):
    """
    Given an array and lookup table name assign the LUT
    to the given array and return the LUT. If aRangeOveride
    is specified then LUT will be mapped through that
    range rather than the array's actual range.
    """
    try:
      self.LUTS[aLutName]
    except KeyError:
      if self.DefaultLUT is not None:
        print ('ERROR: No LUT named %s using %s'%(aLutName,self.DefaultLUT))
        aLutName = self.DefaultLUT
      else:
        print ('ERROR: No LUT named %s and no default available.'%(aLutName))
        return None
    range = self.__GetRange(aArray,aRangeOveride)
    return self.__GetLookupTableForArray(aArray,
               RGBPoints=self.__MapRGB(aLutName,range),
               ColorSpace=self.__GetColorSpace(aLutName),
               VectorMode='Magnitude',
               ScalarRangeInitialized=1.0)

  def GetLUTNames(self):
    """
    Return a list of the currently available LUT's names.
    """
    return sorted(iter(self.LUTS),cmp=lambda x,y: cmp(x.lower(), y.lower()))

  def Print(self):
    """
    Print the available list of LUT's.
    """
    names=""
    i=0
    for k in sorted(iter(self.LUTS),cmp=lambda x,y: cmp(x.lower(), y.lower())):
      lut=self.LUTS[k]
      names+=lut.GetName()
      names+=", "
      if ((i%6)==5):
        names+="\n"
      i+=1
    print (names)
    return
  # end of public interface

  def __GetColorSpace(self,aName):
    """
    Return the color space from the lookup table object.
    """
    return self.LUTS[aName].GetColorSpace()


  def __GetRGB(self,aName):
    """
    Return the rgb values for the named lut
    """
    return self.LUTS[aName]

  def __MapRGB(self,aName,aRange):
    """
    Map the rgb values onto a scalar range
    results are an array of [x r g b] values
    """
    colors=self.LUTS[aName].GetRGBValues()
    mapCoords=self.LUTS[aName].GetMapCoordinates()
    nColors=len(colors)
    coord0=float(aRange[0])
    coordDelta=float(aRange[1])-float(aRange[0])
    mappedColors=[]
    i=0
    while(i<nColors):
      x=coord0+coordDelta*mapCoords[i]
      val=[x]+colors[i]
      mappedColors+=val
      i=i+1
    return mappedColors

  def __GetRange(self,aArray,aRangeOveride):
    """
    Get the range from an array proxy object or if
    an override is provided use that.
    """
    nComps = aArray.GetNumberOfComponents()
    range = [0.0, 1.0]
    if (len(aRangeOveride) == 0):
      if (nComps == 1):
        range = aArray.GetRange()
      else:
        # TODO - this could be larger than the range of the magnitude aArray
        rx = aArray.GetRange(0)
        ry = aArray.GetRange(1)
        rz = aArray.GetRange(2)
        range = [0.0,
             sqrt(rx[1]*rx[1]+ry[1]*ry[1]+rz[1]*rz[1])]
    else:
      range = aRangeOveride
    return range

  def __GetLookupTableForArray(self,aArray,**kwargs):
    """
    Set the lookup table for the given array and assign
    the named properties.
    """
    proxyName='%d.%s.PVLookupTable'%(aArray.GetNumberOfComponents(),aArray.GetName())
    lut = servermanager.ProxyManager().GetProxy('lookup_tables',proxyName)
    if not lut:
      lut = servermanager.rendering.PVLookupTable(ColorSpace="HSV",RGBPoints=[0,0,0,1, 1,1,0,0])
      servermanager.Register(lut, registrationName=proxyName)
    for arg in kwargs.keys():
      if not hasattr(lut, arg):
        raise AttributeError("LUT has no property %s"%(arg))
      setattr(lut,arg,kwargs[arg])
    return lut



r"""
    This module provides some testing functionality for paraview and
    vtk web applications.  It provides the ability to run an arbitrary
    test script in a separate thread and communicate the results back
    to the service so that the CTest framework can be notified of the
    success or failure of the test.

    This test harness will notice when the test script has finished
    running and will notify the service to stop.  At this point, the
    test results will be checked in the main thread which ran the
    service, and in the case of failure an exception will be raised
    to notify CTest of the failure.

    Test scripts need to follow some simple rules in order to work
    within the test harness framework:

    1) implement a function called "runTest(args)", where the args
    parameter contains all the arguments given to the web application
    upon starting.  Among other important items, args will contain the
    port number where the web application is listening.

    2) import the testing module so that the script has access to
    the functions which indicate success and failure.  Also the
    testing module contains convenience functions that might be of
    use to the test scripts.

       from vtk.web import testing

    3) Call the "testPass(testName)" or "testFail(testName)" functions
    from within the runTest(args) function to indicate to the framework
    whether the test passed or failed.

"""

import_warning_info = ""
test_module_comm_queue = None

import vtk

# Try standard Python imports
try :
    import os, re, time, datetime, threading, imp, inspect, Queue, types, io
except :
    import_warning_info += "\nUnable to load at least one basic Python module"

# Image comparison imports
try:
    try:
        from PIL import Image
    except ImportError:
        import Image
    except:
        raise
    import base64
    import itertools
except:
    import_warning_info += "\nUnable to load at least one modules necessary for image comparison"

# Browser testing imports
try :
    import selenium
    from selenium import webdriver
except :
    import_warning_info += "\nUnable to load at least one module necessary for browser tests"

# HTTP imports
try :
    import requests
except :
    import_warning_info += "\nUnable to load at least one module necessary for HTTP tests"


# Define some infrastructure to support different (or no) browsers
test_module_browsers = ["firefox", "chrome", "internet_explorer", "safari", "nobrowser"]
class TestModuleBrowsers:
    firefox, chrome, internet_explorer, safari, nobrowser = range(5)


# =============================================================================
# We can use this exception type to indicate that the test shouldn't actually
# "fail", rather that it was unable to run because some dependencies were not
# met.
# =============================================================================
class DependencyError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


# =============================================================================
# This class allows usage as a dictionary and an object with named property
# access.
# =============================================================================
class Dictionary(dict):
    def __getattribute__(self, attrName):
        return self[attrName]
    def __setattr__(self, attrName, attrValue):
        self[attrName] = attrValue


# =============================================================================
# Checks whether test script supplied, if so, safely imports needed modules
# =============================================================================
def initialize(opts, reactor=None, cleanupMethod=None) :
    """
    This function should be called to initialize the testing module.  The first
    important thing it does is to store the options for later, since the
    startTestThread function will need them.  Then it checks the arguments that
    were passed into the server to see if a test was actually requested, making
    a note of this fact.  Then, if a test was required, this function then
    checks if all the necessary testing modules were safely imported, printing
    a warning if not.  If tests were requested and all modules were present,
    then this function sets "test_module_do_testing" to True and sets up the
    startTestThread function to be called after the reactor is running.

        opts: Parsed arguments from the server

        reactor: This argument is optional, but is used by server.py to
        cause the test thread to be started only after the server itself
        has started.  If it is not provided, the test thread is launched
        immediately.

        cleanupMethod: A callback method you would like the test thread
        to execute when the test has finished.  This is used by server.py
        as a way to have the server terminated after the test has finished,
        but could be used for other cleanup purposes.  This argument is
        also optional.
    """

    global import_warning_info

    global testModuleOptions
    testModuleOptions = Dictionary()

    # Copy the testing options into something we can easily extend
    for arg in vars(opts):
        optValue = getattr(opts, arg)
        testModuleOptions[arg] = optValue

    # If we got one, add the cleanup method to the testing options
    if cleanupMethod:
        testModuleOptions['cleanupMethod'] = cleanupMethod

    # Check if a test was actually requested
    if (testModuleOptions.testScriptPath != "" and testModuleOptions.testScriptPath is not None) :
        # Check if we ran into trouble with any of the testing imports
        if import_warning_info != "" :
            print("WARNING: Some tests may have unmet dependencies")
            print(import_warning_info)

        if reactor is not None :
            # Add startTest callback to the reactor callback queue, so that
            # the test thread gets started after the reactor is running.  Of
            # course this should only happen if everything is good for tests.
            reactor.callWhenRunning(_start_test_thread)
        else :
            # Otherwise, our aim is to start the thread from another process
            # so just call the start method.
            _start_test_thread()


# =============================================================================
# Grab out the command-line arguments needed for by the testing module.
# =============================================================================
def add_arguments(parser) :
    """
    This function retrieves any command-line arguments that the client-side
    tester needs.  In order to run a test, you will typically just need the
    following:

      --run-test-script => This should be the full path to the test script to
      be run.

      --baseline-img-dir => This should be the 'Baseline' directory where the
      baseline images for this test are located.

      --test-use-browser => This should be one of the supported browser types,
      or else 'nobrowser'.  The choices are 'chrome', 'firefox', 'internet_explorer',
      'safari', or 'nobrowser'.
    """

    parser.add_argument("--run-test-script",
                        default="",
                        help="The path to a test script to run",
                        dest="testScriptPath")

    parser.add_argument("--baseline-img-dir",
                        default="",
                        help="The path to the directory containing the web test baseline images",
                        dest="baselineImgDir")

    parser.add_argument("--test-use-browser",
                        default="nobrowser",
                        help="One of 'chrome', 'firefox', 'internet_explorer', 'safari', or 'nobrowser'.",
                        dest="useBrowser")

    parser.add_argument("--temporary-directory",
                        default=".",
                        help="A temporary directory for storing test images and diffs",
                        dest="tmpDirectory")

    parser.add_argument("--test-image-file-name",
                        default="",
                        help="Name of file in which to store generated test image",
                        dest="testImgFile")


# =============================================================================
# Initialize the test client
# =============================================================================
def _start_test_thread() :
    """
    This function checks whether testing is required and if so, sets up a Queue
    for the purpose of communicating with the thread.  then it starts the
    after waiting 5 seconds for the server to have a chance to start up.
    """

    global test_module_comm_queue
    test_module_comm_queue = Queue.Queue()

    t = threading.Thread(target=launch_web_test,
                         args = [],
                         kwargs = { 'serverOpts': testModuleOptions,
                                    'commQueue': test_module_comm_queue,
                                    'testScript': testModuleOptions.testScriptPath })

    t.start()


# =============================================================================
# Test scripts call this function to indicate passage of their test
# =============================================================================
def test_pass(testName) :
    """
    Test scripts should call this function to indicate that the test passed.  A
    note is recorded that the test succeeded, and is checked later on from the
    main thread so that CTest can be notified of this result.
    """

    global test_module_comm_queue
    resultObj = { testName: 'pass' }
    test_module_comm_queue.put(resultObj)


# =============================================================================
# Test scripts call this function to indicate failure of their test
# =============================================================================
def test_fail(testName) :
    """
    Test scripts should call this function to indicate that the test failed.  A
    note is recorded that the test did not succeed, and this note is checked
    later from the main thread so that CTest can be notified of the result.

    The main thread is the only one that can signal test failure in
    CTest framework, and the main thread won't have a chance to check for
    passage or failure of the test until the main loop has terminated.  So
    here we just record the failure result, then we check this result in the
    processTestResults() function, throwing an exception at that point to
    indicate to CTest that the test failed.
    """

    global test_module_comm_queue
    resultObj = { testName: 'fail' }
    test_module_comm_queue.put(resultObj)


# =============================================================================
# Concatenate any number of strings into a single path string.
# =============================================================================
def concat_paths(*pathElts) :
    """
    A very simple convenience function so that test scripts can build platform
    independent paths out of a list of elements, without having to import the
    os module.

        pathElts: Any number of strings which should be concatenated together
        in a platform independent manner.
    """

    return os.path.join(*pathElts)


# =============================================================================
# So we can change our time format in a single place, this function is
# provided.
# =============================================================================
def get_current_time_string() :
    """
    This function returns the current time as a string, using ISO 8601 format.
    """

    return datetime.datetime.now().isoformat(" ")


# =============================================================================
# Uses vtkTesting to compare images.  According to comments in the vtkTesting
# C++ code (and this seems to work), if there are multiple baseline images in
# the same directory as the baseline_img, and they follow the naming pattern:
# 'img.png', 'img_1.png', ... , 'img_N.png', then all of these images will be
# tried for a match.
# =============================================================================
def compare_images(test_img, baseline_img, tmp_dir="."):
    """
    This function creates a vtkTesting object, and specifies the name of the
    baseline image file, using a fully qualified path (baseline_img must be
    fully qualified).  Then it calls the vtkTesting method which compares the
    image (test_img, specified only with a relative path) against the baseline
    image as well as any other images in the same directory as the baseline
    image which follow the naming pattern: 'img.png', 'img_1.png', ... , 'img_N.png'

        test_img: File name of output image to be compared against baseline.

        baseline_img: Fully qualified path to first of the baseline images.

        tmp_dir: Fully qualified path to a temporary directory for storing images.
    """

    # Create a vtkTesting object and specify a baseline image
    t = vtk.vtkTesting()
    t.AddArgument("-T")
    t.AddArgument(tmp_dir)
    t.AddArgument("-V")
    t.AddArgument(baseline_img)

    # Perform the image comparison test and print out the result.
    return t.RegressionTest(test_img, 0.0)


# =============================================================================
# Provide a wait function
# =============================================================================
def wait_with_timeout(delay=None, limit=0, criterion=None):
    """
    This function provides the ability to wait for a certain number of seconds,
    or else to wait for a specific criterion to be met.
    """
    for i in itertools.count():
        if criterion is not None and criterion():
            return True
        elif delay * i > limit:
            return False
        else:
            time.sleep(delay)


# =============================================================================
# Define a WebTest class with five stages of testing: initialization, setup,
# capture, postprocess, and cleanup.
# =============================================================================
class WebTest(object) :
    """
    This is the base class for all automated web-based tests.  It defines five
    stages that any test must run through, and allows any or all of these
    stages to be overridden by subclasses.  This class defines the run_test
    method to invoke the five stages overridden by subclasses, one at a time:
    1) initialize, 2) setup, 3) capture, 4) postprocess, and 5) cleanup.
    """
    class Abort:
        pass

    def __init__(self, url=None, testname=None, **kwargs) :
        self.url = url
        self.testname = testname

    def run_test(self):
        try:
            self.checkdependencies()
            self.initialize()
            self.setup()
            self.capture()
            self.postprocess()
        except WebTest.Abort:
            # Placeholder for future option to return failure result
            pass
        except :
            self.cleanup()
            raise

        self.cleanup()

    def checkdependencies(self):
        pass

    def initialize(self):
        pass

    def setup(self):
        pass

    def capture(self):
        pass

    def postprocess(self):
        pass

    def cleanup(self):
        pass


# =============================================================================
# Define a WebTest subclass designed specifically for browser-based tests.
# =============================================================================
class BrowserBasedWebTest(WebTest):
    """
    This class can be used as a base for any browser-based web tests.  It
    introduces the notion of a selenium browser and overrides phases (1) and
    (3), initialization and cleanup, of the test phases introduced in the base
    class.  Initialization involves selecting the browser type, setting the
    browser window size, and asking the browser to load the url.  Cleanup
    involves closing the browser window.
    """
    def __init__(self, size=None, browser=None, **kwargs):
        self.size = size
        self.browser = browser
        self.window = None

        WebTest.__init__(self, **kwargs)

    def initialize(self):
        try:
            if self.browser is None or self.browser == TestModuleBrowsers.chrome:
                self.window = webdriver.Chrome()
            elif self.browser == TestModuleBrowsers.firefox:
                self.window = webdriver.Firefox()
            elif self.browser == TestModuleBrowsers.internet_explorer:
                self.window = webdriver.Ie()
            else:
                raise DependencyError("self.browser argument has illegal value %r" % (self.browser))
        except DependencyError as dErr:
            raise
        except Exception as inst:
            raise DependencyError(inst)

        if self.size is not None:
            self.window.set_window_size(self.size[0], self.size[1])

        if self.url is not None:
            self.window.get(self.url)

    def cleanup(self):
        try:
            self.window.quit()
        except:
            print('Unable to call window.quit, perhaps this is expected because of unmet browser dependency.')


# =============================================================================
# Extend BrowserBasedWebTest to handle vtk-style image comparison
# =============================================================================
class ImageComparatorWebTest(BrowserBasedWebTest):
    """
    This class extends browser based web tests to include image comparison.  It
    overrides the capture phase of testing with some functionality to simply
    grab a screenshot of the entire browser window.  It overrides the
    postprocess phase with a call to vtk image comparison functionality.
    Derived classes can then simply override the setup function with a series
    of selenium-based browser interactions to create a complete test.  Derived
    classes may also prefer to override the capture phase to capture only
    certain portions of the browser window for image comparison.
    """
    def __init__(self, filename=None, baseline=None, temporaryDir=None, **kwargs):
        if filename is None:
            raise TypeError("missing argument 'filename'")
        if baseline is None:
            raise TypeError("missing argument 'baseline'")

        BrowserBasedWebTest.__init__(self, **kwargs)
        self.filename = filename
        self.baseline = baseline
        self.tmpDir = temporaryDir

    def capture(self):
        self.window.save_screenshot(self.filename)

    def postprocess(self):
        result = compare_images(self.filename, self.baseline, self.tmpDir)

        if result == 1 :
            test_pass(self.testname)
        else :
            test_fail(self.testname)


# =============================================================================
# Given a css selector to use in finding the image element, get the element,
# then base64 decode the "src" attribute and return it.
# =============================================================================
def get_image_data(browser, cssSelector) :
    """
    This function takes a selenium browser and a css selector string and uses
    them to find the target HTML image element.  The desired image element
    should contain it's image data as a Base64 encoded JPEG image string.
    The 'src' attribute of the image is read, Base64-decoded, and then
    returned.

        browser: A selenium browser instance, as created by webdriver.Chrome(),
        for example.

        cssSelector: A string containing a CSS selector which will be used to
        find the HTML image element of interest.
    """

    # Here's maybe a better way to get at that image element
    imageElt = browser.find_element_by_css_selector(cssSelector)

    # Now get the Base64 image string and decode it into image data
    base64String = imageElt.get_attribute("src")
    b64RegEx = re.compile(r'data:image/jpeg;base64,(.+)')
    b64Matcher = b64RegEx.match(base64String)
    imgdata = base64.b64decode(b64Matcher.group(1))

    return imgdata


# =============================================================================
# Combines a variation on above function with the write_image_to_disk function.
# converting jpg to png in the process, if necessary.
# =============================================================================
def save_image_data_as_png(browser, cssSelector, imgfilename) :
    """
    This function takes a selenium browser instance, a css selector string,
    and a file name.  It uses the css selector string to finds the target HTML
    Image element, which should contain a Base64 encoded JPEG image string,
    it decodes the string to image data, and then saves the data to the file.
    The image type of the written file is determined from the extension of the
    provided filename.

        browser: A selenium browser instance as created by webdriver.Chrome(),
        for example.

        cssSelector: A string containing a CSS selector which will be used to
        find the HTML image element of interest.

        imgFilename: The filename to which to save the image. The extension is
        used to determine the type of image which should be saved.
    """
    imageElt = browser.find_element_by_css_selector(cssSelector)
    base64String = imageElt.get_attribute("src")
    b64RegEx = re.compile(r'data:image/jpeg;base64,(.+)')
    b64Matcher = b64RegEx.match(base64String)
    img = Image.open(io.BytesIO(base64.b64decode(b64Matcher.group(1))))
    img.save(imgfilename)


# =============================================================================
# Given a decoded image and the full path to a file, write the image to the
# file.
# =============================================================================
def write_image_to_disk(imgData, filePath) :
    """
    This function takes an image data, as returned by this module's
    get_image_data() function for example, and writes it out to the file given by
    the filePath parameter.

        imgData: An image data object
        filePath: The full path, including the file name and extension, where
        the image should be written.
    """

    with open(filePath, 'wb') as f:
        f.write(imgData)


# =============================================================================
# There could be problems if the script file has more than one class defn which
# is a subclass of vtk.web.testing.WebTest, so we should write some
# documentation to help people avoid that.
# =============================================================================
def instantiate_test_subclass(pathToScript, **kwargs) :
    """
    This function takes the fully qualified path to a test file, along with
    any needed keyword arguments, then dynamically loads the file as a module
    and finds the test class defined inside of it via inspection.  It then
    uses the keyword arguments to instantiate the test class and return the
    instance.

        pathToScript: Fully qualified path to python file containing defined
        subclass of one of the test base classes.
        kwargs: Keyword arguments to be passed to the constructor of the
        testing subclass.
    """

    # Load the file as a module
    moduleName = imp.load_source('dynamicTestModule', pathToScript)
    instance = None

    # Inspect dynamically loaded module members
    for name, obj in inspect.getmembers(moduleName) :
        # Looking for classes only
        if inspect.isclass(obj) :
            instance = obj.__new__(obj)
            # And only classes defined in the dynamically loaded module
            if instance.__module__ == 'dynamicTestModule' :
                try :
                    instance.__init__(**kwargs)
                    break;
                except Exception as inst:
                    print('Caught exception: ' + str(type(inst)))
                    print(inst)
                    raise

    return instance


# =============================================================================
# For testing purposes, define a function which can interact with a running
# paraview or vtk web application service.
# =============================================================================
def launch_web_test(*args, **kwargs) :
    """
    This function loads a python file as a module (with no package), and then
    instantiates the class it must contain, and finally executes the run_test()
    method of the class (which the class may override, but which is defined in
    both of the testing base classes, WebTest and ImageComparatorBaseClass).
    After the run_test() method finishes, this function will stop the web
    server if required.  This function expects some keyword arguments will be
    present in order for it to complete it's task:

        kwargs['serverOpts']: An object containing all the parameters used
        to start the web service.  Some of them will be used in the test script
        in order perform the test.  For example, the port on which the server
        was started will be required in order to connect to the server.

        kwargs['testScript']: The full path to the python file containing the
        testing subclass.
    """

    serverOpts = None
    testScriptFile = None

    # This is really the thing all test scripts will need: access to all
    # the options used to start the server process.
    if 'serverOpts' in kwargs :
        serverOpts = kwargs['serverOpts']
        # print 'These are the serverOpts we got: '
        # print serverOpts

    # Get the full path to the test script
    if 'testScript' in kwargs :
        testScriptFile = kwargs['testScript']

    testName = 'unknown'

    # Check for a test file (python file)
    if testScriptFile is None :
        print('No test script file found, no test script will be run.')
        test_fail(testName)

    # The test name will be generated from the python script name, so
    # match and capture a bunch of contiguous characters which are
    # not '.', '\', or '/', followed immediately by the string '.py'.
    fnamePattern = re.compile('([^\.\/\\\]+)\.py')
    fmatch = re.search(fnamePattern, testScriptFile)
    if fmatch :
        testName = fmatch.group(1)
    else :
        print('Unable to parse testScriptFile (' + str(testScriptfile) + '), no test will be run')
        test_fail(testName)

    # If we successfully got a test name, we are ready to try and run the test
    if testName != 'unknown' :

        # Output file and baseline file names are generated from the test name
        imgFileName = testName + '.png'
        knownGoodFileName = concat_paths(serverOpts.baselineImgDir, imgFileName)
        tempDir = serverOpts.tmpDirectory
        testImgFileName = serverOpts.testImgFile

        testBrowser = test_module_browsers.index(serverOpts.useBrowser)

        # Now try to instantiate and run the test
        try :
            testInstance = instantiate_test_subclass(testScriptFile,
                                                     testname=testName,
                                                     host=serverOpts.host,
                                                     port=serverOpts.port,
                                                     browser=testBrowser,
                                                     filename=testImgFileName,
                                                     baseline=knownGoodFileName,
                                                     temporaryDir=tempDir)

            # If we were able to instantiate the test, run it, otherwise we
            # consider it a failure.
            if testInstance is not None :
                try:
                    testInstance.run_test()
                except DependencyError as derr:
                    # TODO: trigger return SKIP_RETURN_CODE when CMake 3 is required
                    print('Some dependency of this test was not met, allowing it to pass')
                    test_pass(testName)
            else :
                print('Unable to instantiate test instance, failing test')
                test_fail(testName)
                return

        except Exception as inst :
            import sys, traceback
            tb = sys.exc_info()[2]
            print('Caught an exception while running test script:')
            print('  ' + str(type(inst)))
            print('  ' + str(inst))
            print('  ' + ''.join(traceback.format_tb(tb)))
            test_fail(testName)

    # If we were passed a cleanup method to run after testing, invoke it now
    if 'cleanupMethod' in serverOpts :
        serverOpts['cleanupMethod']()


# =============================================================================
# To keep the service module clean, we'll process the test results here, given
# the test result object we generated in "launch_web_test".  It is
# passed back to this function after the service has completed.  Failure of
# of the test is indicated by raising an exception in here.
# =============================================================================
def finalize() :
    """
    This function checks the module's global test_module_comm_queue variable for a
    test result.  If one is found and the result is 'fail', then this function
    raises an exception to communicate the failure to the CTest framework.

    In order for a test result to be found in the test_module_comm_queue variable,
    the test script must have called either the testPass or testFail functions
    provided by this test module before returning.
    """

    global test_module_comm_queue

    if test_module_comm_queue is not None :
        resultObject = test_module_comm_queue.get()

        failedATest = False

        for testName in resultObject :
            testResult = resultObject[testName]
            if testResult == 'fail' :
                print('  Test -> ' + testName + ': ' + testResult)
                failedATest = True

        if failedATest is True :
            raise Exception("At least one of the requested tests failed.  " +
                            "See detailed output, above, for more information")



"""
A module that composites one or more layers together to produce a recognizable
image.
"""

import abc
import numpy as np


class Compositor(object):
    __metaclass__ = abc.ABCMeta
    ''' Base class for different compositor specifications.'''
    def __init__(self):
        self.__bgColor = tuple([0, 0, 0, 0])

    @abc.abstractmethod
    def __renderImpl(self, layers):
        return

    def enableGeometryColor(self, enable):
        # print "Warning: geometry color not supported in this specification."
        return

    def enableLighting(self, enable):
        # print "Warning: lighting not supported in this specification."
        return

    def setColorDefinitions(self, colorDefs):
        return

    def set_background_color(self, rgb):
        self.__bgColor = rgb

    def render(self, layers):
        return self.__renderImpl(layers)


class Compositor_SpecA(Compositor):
    '''
    Compositor SpecA. Loads LayerSpecs consisting of a
    single image. No compositing supported.
    '''
    def __init__(self, parent=None):
        super(Compositor_SpecA, self).__init__()

    def _Compositor__renderImpl(self, layers):
        layer = layers[0] if (len(layers) > 0) else None

        if layer is None:
            raise IndexError("There are no valid layers to render!")

        return layer.getColorArray()


class Compositor_SpecB(Compositor):
    '''
    Compositor SpecB. Composites several LayerSpec instances together.
    Each LayerSpec holds all of its buffer components (depth, color, luminance,
    etc.).
    '''
    def __init__(self, parent=None):
        super(Compositor_SpecB, self).__init__()
        self.__geometryColorEnabled = False
        self.__lightingEnabled = True
        self.__colorDefinitions = {}

    def ambient(self, rgb):
        """ Returns the ambient contribution in an RGB luminance image. """
        return np.dstack((rgb[:, :, 0], rgb[:, :, 0], rgb[:, :, 0]))

    def diffuse(self, rgb):
        """ Returns the diffuse contribution in an RGB luminance image. """
        return np.dstack((rgb[:, :, 1], rgb[:, :, 1], rgb[:, :, 1]))

    def specular(self, rgb):
        """ Returns the specular contribution in an RGB luminance image. """
        return np.dstack((rgb[:, :, 2], rgb[:, :, 2], rgb[:, :, 2]))

    def enableGeometryColor(self, enable):
        self.__geometryColorEnabled = enable

    def enableLighting(self, enable):
        self.__lightingEnabled = enable

    def setColorDefinitions(self, colorDefs):
        self.__colorDefinitions = colorDefs

    def __getCustomizedColorBuffer(self, layer):
        '''
        Queries the user-defined color customizations from a dictionary and
        applies them to a specific LayerSpec instance. The color
        customizations form the ui are matched to its referred LayerSpec using
        the layer name (a 'parameter/value' tag), which is set in the
        QueryTranslator. See the PipelineModel for more information on the
        colorDefinitions struct.
        '''
        array = None
        customizationName = layer.customizationName
        if layer.hasValueArray():
            array = layer.getValueArray()
            valueRange = layer.valueRange

            if customizationName in self.__colorDefinitions:
                array = self.__applyColorLut(
                    array, layer.getDepth(),
                    self.__colorDefinitions[customizationName]["colorLut"],
                    valueRange)

        elif layer.hasColorArray():
            array = np.copy(layer.getColorArray())

            if customizationName in self.__colorDefinitions:
                self.__applyFillColor(
                    array, layer.getDepth(),
                    self.__colorDefinitions[customizationName])

        return array

    def __applyFillColor(self, array, depth, colorDef):
        ''' Applies the user defined geometry color to an array.
        TODO Performance is notoriously affected when calling this function.
        Needs optimization. '''
        if self.__geometryColorEnabled:
            # Process only foreground (object) values
            arrIdx = self.__getForegroundPixels(depth)
            if not arrIdx:
                return
            fill_color = colorDef["geometryColor"]
            array[arrIdx[0], arrIdx[1], :] = fill_color[0:3]

    def __applyColorLut(self, rgbVarr, depth, colorLutStruct, valueRange):
        if colorLutStruct.name == "None":
            return rgbVarr
        # Process only foreground (object) values
        varrIdx = self.__getForegroundPixels(depth)

        if len(rgbVarr.shape) == 2 and rgbVarr.dtype == np.float32:
            if not varrIdx:
                # No foreground. Return a dummy 3c-uint8 image
                # (returning a single chan float image causes issues at
                # the PIL conversion)
                return np.zeros((rgbVarr.shape[0], rgbVarr.shape[1], 3),
                                np.uint8)

            return self.__floatToRGB(
                rgbVarr, varrIdx, colorLutStruct, valueRange)
        else:
            if not varrIdx:
                return rgbVarr

            return self.__invertibleToRGB(rgbVarr, varrIdx, colorLutStruct)

    def __floatToRGB(self, rgbVarr, varrIdx, colorLutStruct, valueRange):
        ''' Decode a value image rendered in FLOATING_POINT mode. The image
        is treated as a plain float buffer, so the color table can be applied
        directly.'''
        # Normalized foreground values down to 0..1 to use as LUT indexes

        # what do we have from raster itself?
        foreground = rgbVarr[varrIdx[0], varrIdx[1]]
        valueMin = foreground.min()
        valueMax = foreground.max()

        if (valueMin >= 0.0) and (valueMax <= 1.0):
            # most likely came from an RGB render, in which case
            # already in 0..1
            # todo: grab from store.metadata.value_mode !=2 instead
            pass
        else:
            # from a direct value render, numbers are actual
            if valueRange:
                # we were given a global range, use it instead of local
                # range from just this slice
                valueMin = valueRange[0]
                valueMax = valueRange[1]
                # this provides a way to clamp, for example NaN's
                foreground[foreground < valueMin] = valueMin
                foreground[foreground > valueMax] = valueMax
            valRange = valueMax - valueMin
            if valRange > 1e-4:
                foreground = (foreground - valueMin) / valRange
            else:
                # no way to know where this single value lies, so use bottom
                foreground.fill(0.0)

        colorLut = colorLutStruct.lut
        bins = colorLutStruct.adjustedBins

        colorIndices = np.digitize(foreground, bins)
        colorIndices = np.subtract(colorIndices, 1)

        shape = rgbVarr.shape
        valueImage = np.zeros([shape[0], shape[1]], dtype=np.uint32)
        valueImage[varrIdx[0], varrIdx[1]] = colorIndices

        return colorLut[valueImage]

    def __invertibleToRGB(self, rgbVarr, varrIdx, colorLutStruct):
        '''
        Decode an RGB image rendered in INVERTIBLE_LUT mode. The image encodes
        float values as colors, so the RGB value is first decoded into its
        represented float value and then the color table is applied.
        '''
        w0 = np.left_shift(
            rgbVarr[varrIdx[0], varrIdx[1], 0].astype(np.uint32), 16)
        w1 = np.left_shift(
            rgbVarr[varrIdx[0], varrIdx[1], 1].astype(np.uint32), 8)
        w2 = rgbVarr[varrIdx[0], varrIdx[1], 2]

        value = np.bitwise_or(w0, w1)
        value = np.bitwise_or(value, w2)
        value = np.subtract(value.astype(np.int32), 1)
        normalized_val = np.divide(value.astype(float), 0xFFFFFE)

        # Map float value to color lut (use a histogram to support non-uniform
        # colormaps (fetch bin indices))
        colorLut = colorLutStruct.lut
        bins = colorLutStruct.adjustedBins

        idx = np.digitize(normalized_val, bins)
        idx = np.subtract(idx, 1)

        valueImage = np.zeros([rgbVarr.shape[0], rgbVarr.shape[1]],
                              dtype=np.uint32)
        valueImage[varrIdx[0], varrIdx[1]] = idx

        return colorLut[valueImage]

    def __getForegroundPixels(self, depth):
        '''
        Computes the indices of the foreground object using the depth buffer.
        '''
        bgDepth = np.max(depth)
        indices = np.where(depth < bgDepth)
        if len(indices[0]) == 0:
            # Only background
            return None
        return indices

    def _Compositor__renderImpl(self, layers):
        """
        Takes an array of layers (LayerSpec) and composites them into an RGB
        image.
        """
        # find a valid LayerSpec (valid = at least one color loaded)
        # necessary for compatibility SpecB-testcase1
        l0 = None
        for index, layer in enumerate(layers):
            if layer.hasColorArray() or layer.hasValueArray():
                # TODO might be necessary to check if the actual array is
                # not NONE too
                l0 = layer
                break

        if l0 is None:
            raise IndexError("There are no valid layers to render!")

        c0 = self.__getCustomizedColorBuffer(l0)

        if self.__lightingEnabled:
            lum0 = l0.getLuminance()
            if lum0 is not None:
                # modulate color of first layer by the luminance
                lum0 = np.copy(lum0)
                lum0 = self.diffuse(lum0)
                c0[:, :, :] = c0[:, :, :] * (lum0[:, :, :]/255.0)

        # composite the rest of the layers
        d0 = np.copy(l0.getDepth())
        for idx in range(1, len(layers)):

            cnext = self.__getCustomizedColorBuffer(layers[idx])
            # necessary for compatibility Spect-testcase1
            if cnext is None:
                continue

            dnext = layers[idx].getDepth()
            lnext = layers[idx].getLuminance()

            # put the top pixels into place
            indices = np.where(dnext < d0)
            if (self.__lightingEnabled and lnext is not None):
                # modulate color by luminance then insert
                lnext = self.diffuse(lnext)
                c0[indices[0], indices[1], :] = \
                    cnext[indices[0], indices[1], :] * \
                    (lnext[indices[0], indices[1], :] / 255.0)
            else:
                # no luminance, direct insert
                c0[indices[0], indices[1], :] = cnext[
                    indices[0], indices[1], :]

            d0[indices[0], indices[1]] = dnext[indices[0], indices[1]]

        if not (d0 is None) and not (d0.ndim == 0):
            # set background pixels to gray to avoid colormap
            # TODO: curious why necessary, we encode a NaN value on these
            # pixels?
            indices = np.where(d0 == np.max(d0))
            __bgColor = self._Compositor__bgColor
            c0[indices[0], indices[1], 0] = __bgColor[0]
            c0[indices[0], indices[1], 1] = __bgColor[1]
            c0[indices[0], indices[1], 2] = __bgColor[2]

        return c0



from __future__ import absolute_import
from . import vtkPVClientServerCoreCore
from . import vtkPVVTKExtensionsDefault
from . import vtkPVVTKExtensionsRendering
from vtkmodules import vtkDomainsChemistry
from vtkmodules import vtkFiltersAMR
from vtkmodules import vtkRenderingLabel
from vtkmodules import vtkRenderingVolumeOpenGL2
from vtkmodules import vtkViewsContext2D
from vtkmodules import vtkViewsCore
from .vtkPVClientServerCoreRenderingPython import *



from __future__ import absolute_import
from . import vtkCommonCore
from . import vtkCommonDataModel
from . import vtkCommonExecutionModel
from . import vtkIOLegacy
from . import vtkIOXMLParser
from . import vtkRenderingCore
from .vtkDomainsChemistryPython import *



# Macro for BUG #11065. This makes it possible to show the grid for a dataset in
# the background.
try: paraview.simple
except: from paraview.simple import *
paraview.simple._DisableFirstRenderCameraReset()

spcth_0 = GetActiveSource()
ExtractSurface2 = ExtractSurface()
DataRepresentation5 = Show()
DataRepresentation5.Representation = 'Wireframe'
DataRepresentation5.BackfaceRepresentation = 'Cull Frontface'



from __future__ import absolute_import
from . import vtkCommonDataModel
from . import vtkCommonExecutionModel
from . import vtkCommonMisc
from . import vtkFiltersCore
from . import vtkFiltersGeneral
from .vtkFiltersHyperTreePython import *



from __future__ import absolute_import
from . import vtkPVClientServerCoreRendering
from . import vtkPVVTKExtensionsDefault
from vtkmodules import vtkFiltersProgrammable
from .vtkPVClientServerCoreDefaultPython import *



r"""python_view is a module providing access to a PythonView. It is
possible to use the PythonView API directly, but this module provides
convenience classes in Python.
"""
#==============================================================================
#
#  Program:   ParaView
#  Module:    python_view.py
#
#  Copyright (c) Kitware, Inc.
#  All rights reserved.
#  See Copyright.txt or http://www.paraview.org/HTML/Copyright.html for details.
#
#     This software is distributed WITHOUT ANY WARRANTY without even
#     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#     PURPOSE.  See the above copyright notice for more information.
#
#==============================================================================
import paraview
from vtkmodules.vtkCommonDataModel import vtkImageData

def numpy_to_image(numpy_array):
  """Convert a numpy 2D or 3D array to a vtkImageData object.

  numpy_array
    2D or 3D numpy array containing image data

  return
    vtkImageData with the numpy_array content
  """
  try:
    import numpy
  except:
    paraview.print_error("Error: Cannot import numpy")

  shape = numpy_array.shape
  if len(shape) < 2:
    raise Exception('numpy array must have dimensionality of at least 2')

  h, w = shape[0], shape[1]
  c = 1
  if len(shape) == 3:
    c = shape[2]

  # Reshape 2D image to 1D array suitable for conversion to a
  # vtkArray with numpy_support.numpy_to_vtk()
  linear_array = numpy.reshape(numpy_array, (w*h, c))

  try:
    from vtkmodules.util import numpy_support
  except:
    paraview.print_error("Error: Cannot import vtkmodules.util.numpy_support")

  vtk_array = numpy_support.numpy_to_vtk(linear_array)

  image = vtkImageData()
  image.SetDimensions(w, h, 1)
  image.AllocateScalars(vtk_array.GetDataType(), 4)
  image.GetPointData().GetScalars().DeepCopy(vtk_array)

  return image


def figure_to_data(figure):
  """Convert a Matplotlib figure to a numpy 2D array with RGBA uint8 channels
  and return it.

  figure
    A matplotlib figure.

  return
    A numpy 2D array of RGBA values.
  """
  # Draw the renderer
  try:
    import matplotlib
  except:
    paraview.print_error("Error: Cannot import matplotlib")

  figure.canvas.draw()

  # Get the RGBA buffer from the figure
  w, h = figure.canvas.get_width_height()

  try:
    import numpy
  except:
    paraview.print_error("Error: Cannot import numpy")

  buf = numpy.fromstring(figure.canvas.tostring_argb(), dtype=numpy.uint8)
  buf.shape = (h, w, 4)

  # canvas.tostring_argb gives pixmap in ARGB mode. Roll the alpha channel to have it in RGBA mode
  buf = numpy.roll(buf, 3, axis=2)

  return buf


def figure_to_image(figure):
  """Convert a Matplotlib figure to a vtkImageData with RGBA unsigned char
  channels.

  figure
    A matplotlib figure.

  return
    A vtkImageData with the Matplotlib figure content.
  """
  buf = figure_to_data(figure)

  # Flip rows to be suitable for vtkImageData.
  buf = buf[::-1,:,:].copy()

  return numpy_to_image(buf)


def matplotlib_figure(width, height):
  """Create a Matplotlib figure with specified width and height for rendering.

  w
    Width of desired plot.
  h
    Height of desired plot.

  return
    A Matplotlib figure.
  """
  try:
    from matplotlib.backends.backend_agg import FigureCanvasAgg
  except:
    paraview.print_error("Error: Cannot import matplotlib.backends.backend_agg.FigureCanvasAgg")
  try:
    from matplotlib.figure import Figure
  except:
    paraview.print_error("Error: Cannot import matplotlib.figure.Figure")

  figure = Figure()
  figureCanvas = FigureCanvasAgg(figure)
  figure.set_dpi(72)
  figure.set_size_inches(float(width)/72.0, float(height)/72.0)

  return figure


def call_render(render_function, view, width, height):
  """Utility function to call the user-defined render function. This is called
  by the C++ side of the vtkPythonView class.

  view
    vtkPythonView object

  width
    Width of view

  height
    Height of view
  """
  if render_function == None:
    return

  # Check how many parameters it takes.
  num_args = render_function.__code__.co_argcount

  image = None
  if (num_args == 3):
    # Current-style render() function
    image = render_function(view, width, height)

  elif (num_args == 2):
    # Old-style render() function introduced in ParaView 4.1
    figure = matplotlib_figure(width, height)
    render_function(view, figure)

    image = figure_to_image(figure)

  view.SetImageData(image)



r"""Module for generating a Python state for ParaView.
This module uses paraview.smtrace to generate a trace for a selected set of
proxies my mimicking the creating of various pipeline components in sequence.
Typical usage of this module is as follows::

    from paraview import smstate
    state = smstate.get_state()
    print (state)

Note, this cannot be called when Python tracing is active.
"""

from paraview import servermanager as sm
from paraview import smtrace
from paraview import simple
import sys

if sys.version_info >= (3,):
    xrange = range

class supported_proxies(object):
    """filter object used to hide proxies that are currently not supported by
    the state saving mechanism or those that are generally skipped in state e.g.
    animation proxies and time keeper."""
    def __call__(self, proxy):
        return proxy and \
            not proxy.GetXMLGroup() == "animation" and \
            not proxy.GetXMLName() == "TimeKeeper"

class visible_representations(object):
    """filter object to skip hidden representations from being saved in state file"""
    def __call__(self, proxy):
        if not supported_proxies()(proxy): return False
        try:
            return proxy.Visibility
        except AttributeError:
            pass
        return True

def __toposort(input_set):
    """implementation of Tarjan topological sort to sort proxies using consumer
    dependencies as graph edges."""
    result = []
    marked_set = set()
    while marked_set != input_set:
        unmarked_node = (input_set - marked_set).pop()
        __toposort_visit(result, unmarked_node, input_set, marked_set)
    result.reverse()
    return result

def __toposort_visit(result, proxy, input_set, marked_set, t_marked_set=None):
    if t_marked_set is None:
        temporarily_marked_set = set()
    else:
        temporarily_marked_set = t_marked_set
    if proxy in temporarily_marked_set:
        raise RuntimeError ("Cycle detected in pipeline! %r" % proxy)
    if not proxy in marked_set:
        temporarily_marked_set.add(proxy)
        consumers = set()
        get_consumers(proxy, lambda x: x in input_set, consumer_set=consumers, recursive=False)
        for x in consumers:
            __toposort_visit(result, x, input_set, marked_set, temporarily_marked_set)
        marked_set.add(proxy)
        temporarily_marked_set.discard(proxy)
        result.append(proxy)

def get_consumers(proxy, filter, consumer_set, recursive=True):
    """Returns the consumers for a proxy iteratively. If filter is non-None,
    filter is used to cull consumers."""
    for i in xrange(proxy.GetNumberOfConsumers()):
        consumer = proxy.GetConsumerProxy(i)
        consumer = consumer.GetTrueParentProxy() if consumer else None
        consumer = sm._getPyProxy(consumer)
        if not consumer or consumer.IsPrototype() or consumer in consumer_set:
            continue
        if filter(consumer):
            consumer_set.add(consumer)
            if recursive: get_consumers(consumer, filter, consumer_set)

def get_producers(proxy, filter, producer_set):
    """Returns the producers for a proxy iteratively. If filter is non-None,
    filter is used to cull producers."""
    for i in xrange(proxy.GetNumberOfProducers()):
        producer = proxy.GetProducerProxy(i)
        producer = producer.GetTrueParentProxy() if producer else None
        producer = sm._getPyProxy(producer)
        if not producer or producer.IsPrototype() or producer in producer_set:
            continue
        if filter(producer):
            producer_set.add(producer)
            get_producers(producer, filter, producer_set)
    # FIXME: LookupTable is missed :/, darn subproxies!
    try:
        if proxy.LookupTable and filter(proxy.LookupTable):
            producer_set.add(proxy.LookupTable)
            get_producers(proxy.LookupTable, filter, producer_set)
    except AttributeError: pass
    try:
        if proxy.ScalarOpacityFunction and filter(proxy.ScalarOpacityFunction):
            producer_set.add(proxy.ScalarOpacityFunction)
            get_producers(proxy.ScalarOpacityFunction, filter, producer_set)
    except AttributeError: pass

def get_state(propertiesToTraceOnCreate=1, # sm.vtkSMTrace.RECORD_MODIFIED_PROPERTIES,
    skipHiddenRepresentations=True, source_set=[], filter=None, raw=False):
    """Returns the state string"""

    # essential to ensure any obsolete accessor don't linger can cause havoc
    # when saving state following a Python trace session
    # (paraview/paraview#18994)
    import gc
    gc.collect()

    if sm.vtkSMTrace.GetActiveTracer():
        raise RuntimeError ("Cannot generate Python state when tracing is active.")

    if filter is None:
        filter = visible_representations() if skipHiddenRepresentations else supported_proxies()

    # build a set of proxies of interest
    if source_set:
        start_set = source_set
    else:
        # if nothing is specified, we save all views and sources.
        start_set = [x for x in simple.GetSources().values()] + simple.GetViews()
    start_set = [x for x in start_set if filter(x)]

    # now, locate dependencies for the start_set, pruning irrelevant branches
    consumers = set(start_set)
    for proxy in start_set:
        get_consumers(proxy, filter, consumers)

    producers = set()
    for proxy in consumers:
        get_producers(proxy, filter, producers)

    # proxies_of_interest is set of all proxies that we should trace.
    proxies_of_interest = producers.union(consumers)
    #print ("proxies_of_interest", proxies_of_interest)

    trace_config = smtrace.start_trace()
    # this ensures that lookup tables/scalar bars etc. are fully traced.
    trace_config.SetFullyTraceSupplementalProxies(True)

    trace = smtrace.TraceOutput()
    trace.append("# state file generated using %s" % simple.GetParaViewSourceVersion())

    #--------------------------------------------------------------------------
    # First, we trace the views and layouts, if any.
    views = [x for x in proxies_of_interest if smtrace.Trace.get_registered_name(x, "views")]

    if views:
        # sort views by their names, so the state has some structure to it.
        views = sorted(views, key=lambda x:\
                smtrace.Trace.get_registered_name(x, "views"))
        trace.append_separated([\
            "# ----------------------------------------------------------------",
            "# setup views used in the visualization",
            "# ----------------------------------------------------------------"])
        for view in views:
            # FIXME: save view camera positions and size.
            traceitem = smtrace.RegisterViewProxy(view)
            traceitem.finalize()
            del traceitem
        trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))
        trace.append_separated(["SetActiveView(None)"])

    # from views,  build the list of layouts of interest.
    layouts = set()
    for aview in views:
        l = simple.GetLayout(aview)
        if l:
            layouts.add(simple.GetLayout(aview))

    # trace create of layouts
    if layouts:
        layouts = sorted(layouts, key=lambda x:\
                  smtrace.Trace.get_registered_name(x, "layouts"))
        trace.append_separated([\
            "# ----------------------------------------------------------------",
            "# setup view layouts",
            "# ----------------------------------------------------------------"])
        for layout in layouts:
            traceitem = smtrace.RegisterLayoutProxy(layout)
            traceitem.finalize(filter=lambda x: x in views)
            del traceitem
        trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

    if views:
        # restore the active view after the layouts have been created.
        trace.append_separated([\
            "# ----------------------------------------------------------------",
            "# restore active view",
            "SetActiveView(%s)" % smtrace.Trace.get_accessor(simple.GetActiveView()),
            "# ----------------------------------------------------------------"])

    #--------------------------------------------------------------------------
    # Next, trace data processing pipelines.
    sorted_proxies_of_interest = __toposort(proxies_of_interest)
    sorted_sources = [x for x in sorted_proxies_of_interest \
        if smtrace.Trace.get_registered_name(x, "sources")]
    if sorted_sources:
        trace.append_separated([\
            "# ----------------------------------------------------------------",
            "# setup the data processing pipelines",
            "# ----------------------------------------------------------------"])
        for source in sorted_sources:
            traceitem = smtrace.RegisterPipelineProxy(source)
            traceitem.finalize()
            del traceitem
        trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

    #--------------------------------------------------------------------------
    # Can't decide if the representations should be saved with the pipeline
    # objects or afterwords, opting for afterwords for now since the topological
    # sort doesn't guarantee that the representations will follow their sources
    # anyways.
    sorted_representations = [x for x in sorted_proxies_of_interest \
        if smtrace.Trace.get_registered_name(x, "representations")]
    scalarbar_representations = [x for x in sorted_proxies_of_interest\
        if smtrace.Trace.get_registered_name(x, "scalar_bars")]
    # print ("sorted_representations", sorted_representations)
    # print ("scalarbar_representations", scalarbar_representations)
    if sorted_representations or scalarbar_representations:
        for view in views:
            view_representations = [x for x in view.Representations if x in sorted_representations]
            view_scalarbars = [x for x in view.Representations if x in scalarbar_representations]
            if view_representations or view_scalarbars:
                trace.append_separated([\
                    "# ----------------------------------------------------------------",
                    "# setup the visualization in view '%s'" % smtrace.Trace.get_accessor(view),
                    "# ----------------------------------------------------------------"])
            for rep in view_representations:
                try:
                    producer = rep.Input
                    port = rep.Input.Port
                    traceitem = smtrace.Show(producer, port, view, rep,
                        comment="show data from %s" % smtrace.Trace.get_accessor(producer))
                    traceitem.finalize()
                    del traceitem
                    trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

                    if rep.UseSeparateColorMap:
                        trace.append_separated([\
                            "# set separate color map",
                            "%s.UseSeparateColorMap = True" % (\
                                smtrace.Trace.get_accessor(rep))])

                except AttributeError: pass
            # save the scalar bar properties themselves.
            if view_scalarbars:
                trace.append_separated("# setup the color legend parameters for each legend in this view")
                for rep in view_scalarbars:
                    smtrace.Trace.get_accessor(rep)
                    trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))
                    trace.append_separated([\
                      "# set color bar visibility", "%s.Visibility = %s" % (\
                    smtrace.Trace.get_accessor(rep), rep.Visibility)])


            for rep in view_representations:
                try:
                    producer = rep.Input
                    port = rep.Input.Port

                    if rep.IsScalarBarVisible(view):
                        # FIXME: this will save this multiple times, right now,
                        # if two representations use the same LUT.
                        trace.append_separated([\
                            "# show color legend",
                            "%s.SetScalarBarVisibility(%s, True)" % (\
                                smtrace.Trace.get_accessor(rep),
                                smtrace.Trace.get_accessor(view))])

                    if not rep.Visibility:
                      traceitem = smtrace.Hide(producer, port, view)
                      traceitem.finalize()
                      del traceitem
                      trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

                except AttributeError: pass

    #--------------------------------------------------------------------------
    # Now, trace the transfer functions (color maps and opacity maps) used.
    ctfs = set([x for x in proxies_of_interest \
        if smtrace.Trace.get_registered_name(x, "lookup_tables")])
    if ctfs:
        trace.append_separated([\
            "# ----------------------------------------------------------------",
            "# setup color maps and opacity mapes used in the visualization",
            "# note: the Get..() functions create a new object, if needed",
            "# ----------------------------------------------------------------"])
        for ctf in ctfs:
            smtrace.Trace.get_accessor(ctf)
            if ctf.ScalarOpacityFunction in proxies_of_interest:
                smtrace.Trace.get_accessor(ctf.ScalarOpacityFunction)
        trace.append_separated(smtrace.get_current_trace_output_and_reset(raw=True))

    # restore the active source since the order in which the pipeline is created
    # in the state file can end up changing the active source to be different
    # than what it was when the state is being saved.
    trace.append_separated([\
            "# ----------------------------------------------------------------",
            "# finally, restore active source",
            "SetActiveSource(%s)" % smtrace.Trace.get_accessor(simple.GetActiveSource()),
            "# ----------------------------------------------------------------"])
    del trace_config
    smtrace.stop_trace()
    #print (trace)
    return str(trace) if not raw else trace.raw_data()

if __name__ == "__main__":
    print ( "Running test")
    simple.Mandelbrot()
    simple.Show()
    simple.Hide()
    simple.Shrink().ShrinkFactor  = 0.4
    simple.UpdatePipeline()
    simple.Clip().ClipType.Normal[1] = 1

    rep = simple.Show()
    view = simple.Render()
    view.ViewSize=[500, 500]
    rep.SetScalarBarVisibility(view, True)
    simple.Render()
#    rep.SetScalarBarVisibility(view, False)

    print ("====================================================================")
    print (get_state())



r'''
This package provides modules used by ParaView internally and not meant for
public consumption.

This package was added after several internal modules had been directly added to
the `paraview` package. Those modules should eventually be moved to this package
to avoid clobbering the `paraview` namespace.

Modules in this package have no backwards compatibility assurances and will
change without notice.
'''



from __future__ import absolute_import
from . import vtkCommonCore
from . import vtkCommonDataModel
from . import vtkCommonExecutionModel
from . import vtkCommonColor
from . import vtkIOImage
from . import vtkImagingCore
from . import vtkImagingSources
from . import vtkRenderingFreeType
from .vtkInfovisCorePython import *
